@@ -5,6 +5,7 @@ from contextlib import suppress
 from functools import wraps
 from inspect import Signature
 from inspect import _empty as empty_param
+from types import LambdaType
 from typing import (
     TYPE_CHECKING,
     Any,
@@ -25,7 +26,7 @@ from typing import (
 
 from typing_extensions import Self, TypeVar, deprecated
 
-from .builtins import F0, F1, P0, P1, R0, R1, T0, KwargsT, P, R, R0_co, R1_co, R_co, T, T0_co, T1_co, T_co
+from .builtins import F0, F1, P0, P1, R0, T0, KwargsT, P, R, R0_co, R1_co, R_co, T, T0_co, T1_co, T_co
 
 __all__ = [
     "KwargsNotNone",
@@ -401,35 +402,37 @@ def get_subclasses(family: type[T], exclude: Sequence[type[T]] = []) -> list[typ
     return list(set(_subclasses(family)))
 
 
-class classproperty_base(Generic[T, R_co]):
+_T_Any = TypeVar("_T_Any", default=Any)
+_T0_Any = TypeVar("_T0_Any", default=Any)
+
+
+class classproperty_base(Generic[T, R_co, _T_Any]):
     __isabstractmethod__: bool = False
 
+    fget: Callable[[type[T]], R_co]
+    fset: Callable[Concatenate[type[T], _T_Any, ...], None] | None
+    fdel: Callable[[type[T]], None] | None
+
     def __init__(
         self,
         fget: Callable[[type[T]], R_co] | classmethod[T, ..., R_co],
-        fset: Callable[Concatenate[type[T], Any, ...], None]
-        | classmethod[T, Concatenate[Any, ...], None]
+        fset: Callable[Concatenate[type[T], _T_Any, ...], None]
+        | classmethod[T, Concatenate[_T_Any, ...], None]
         | None = None,
         fdel: Callable[[type[T]], None] | classmethod[T, ..., None] | None = None,
         doc: str | None = None,
     ) -> None:
-        self.fget = self._wrap(fget)
-        self.fset = self._wrap(fset) if fset is not None else fset
-        self.fdel = self._wrap(fdel) if fdel is not None else fdel
+        self.fget = fget.__func__ if isinstance(fget, classmethod) else fget
+        self.fset = fset.__func__ if isinstance(fset, classmethod) else fset
+        self.fdel = fdel.__func__ if isinstance(fdel, classmethod) else fdel
 
         self.__doc__ = doc
         self.__name__ = self.fget.__name__
 
-    def _wrap(self, func: Callable[..., R1] | classmethod[T, P1, R1]) -> classmethod[T, P1, R1]:
-        if not isinstance(func, classmethod):
-            func = classmethod(func)
-
-        return func
-
     def __set_name__(self, owner: object, name: str) -> None:
         self.__name__ = name
 
-    def _get_cache(self, type_: type) -> dict[str, Any]:
+    def _get_cache(self, type_: type[T]) -> dict[str, Any]:
         cache_key = getattr(self, "cache_key")
 
         if not hasattr(type_, cache_key):
@@ -437,21 +440,23 @@ class classproperty_base(Generic[T, R_co]):
 
         return getattr(type_, cache_key)
 
-    def __get__(self, obj: T | None, type_: type | None = None) -> R_co:
-        if type_ is None:
+    def __get__(self, obj: T | None, type_: type[T] | None = None) -> R_co:
+        if type_ is None and obj is not None:
             type_ = type(obj)
+        elif type_ is None:
+            raise NotImplementedError("Both obj and type_ are None")
 
         if not isinstance(self, classproperty.cached):
-            return self.fget.__get__(obj, type_)()
+            return self.fget(type_)
 
         if self.__name__ in (cache := self._get_cache(type_)):
             return cache[self.__name__]
 
-        value = self.fget.__get__(obj, type_)()
+        value = self.fget(type_)
         cache[self.__name__] = value
         return value
 
-    def __set__(self, obj: T, value: Any) -> None:
+    def __set__(self, obj: T, value: _T_Any) -> None:
         if not self.fset:
             raise AttributeError(
                 f'classproperty with getter "{self.__name__}" of "{obj.__class__.__name__}" object has no setter.'
@@ -459,10 +464,13 @@ class classproperty_base(Generic[T, R_co]):
 
         type_ = type(obj)
 
+        if not isinstance(self, classproperty.cached):
+            return self.fset(type_, value)
+
         if self.__name__ in (cache := self._get_cache(type_)):
             del cache[self.__name__]
 
-        self.fset.__get__(None, type_)(value)
+        self.fset(type_, value)
 
     def __delete__(self, obj: T) -> None:
         if not self.fdel:
@@ -472,18 +480,21 @@ class classproperty_base(Generic[T, R_co]):
 
         type_ = type(obj)
 
+        if not isinstance(self, classproperty.cached):
+            return self.fdel(type_)
+
         if self.__name__ in (cache := self._get_cache(type_)):
             del cache[self.__name__]
 
-        self.fdel.__get__(None, type_)()
+        self.fdel(type_)
 
 
-class classproperty(classproperty_base[T, R_co]):
+class classproperty(classproperty_base[T, R_co, _T_Any]):
     """
     A combination of `classmethod` and `property`.
     """
 
-    class cached(classproperty_base[T0, R0_co]):
+    class cached(classproperty_base[T0, R0_co, _T0_Any]):
         """
         A combination of `classmethod` and `property`.
 
@@ -516,10 +527,7 @@ class classproperty(classproperty_base[T, R_co]):
                     del cache[name]
 
 
-_T_cc = TypeVar("_T_cc", default=Any)
-
-
-class cachedproperty(property, Generic[R_co, _T_cc]):
+class cachedproperty(property, Generic[R_co, _T_Any]):
     """
     Wrapper for a one-time get property, that will be cached.
 
@@ -541,23 +549,32 @@ class cachedproperty(property, Generic[R_co, _T_cc]):
         def __init__(
             self,
             fget: Callable[[Any], R_co],
-            fset: Callable[[Any, _T_cc], None] | None = None,
+            fset: Callable[[Any, _T_Any], None] | None = None,
             fdel: Callable[[Any], None] | None = None,
             doc: str | None = None,
         ) -> None: ...
 
-        def getter(self, fget: Callable[..., R_co]) -> cachedproperty[R_co, _T_cc]: ...
+        def getter(self, fget: Callable[..., R_co]) -> cachedproperty[R_co, _T_Any]: ...
 
-        def setter(self, fset: Callable[[Any, _T_cc], None]) -> cachedproperty[R_co, _T_cc]: ...
+        def setter(self, fset: Callable[[Any, _T_Any], None]) -> cachedproperty[R_co, _T_Any]: ...
 
-        def deleter(self, fdel: Callable[..., None]) -> cachedproperty[R_co, _T_cc]: ...
+        def deleter(self, fdel: Callable[..., None]) -> cachedproperty[R_co, _T_Any]: ...
 
     if sys.version_info < (3, 13):
 
-        def __init__(self, fget: Any, fset: Any | None = None, fdel: Any | None = None, doc: str | None = None) -> None:
-            self.__name__ = fget.__name__
+        def __init__(
+            self,
+            fget: Callable[[Any], R_co],
+            fset: Callable[[Any, _T_Any], None] | None = None,
+            fdel: Callable[[Any], None] | None = None,
+            doc: str | None = None,
+        ) -> None:
+            self.__name__ = fget.__name__ + f"_{id(fget)}" if isinstance(fget, LambdaType) else fget.__name__
             super().__init__(fget, fset, fdel, doc)
 
+        def __set_name__(self, owner: object, name: str) -> None:
+            self.__name__ = name
+
     @overload
     def __get__(self, instance: None, owner: type | None = None) -> Self: ...
 
@@ -575,7 +592,7 @@ class cachedproperty(property, Generic[R_co, _T_cc]):
         cache[self.__name__] = value
         return value
 
-    def __set__(self, instance: Any, value: _T_cc) -> None:
+    def __set__(self, instance: Any, value: _T_Any) -> None:
         if self.__name__ in (cache := instance.__dict__.setdefault(self.cache_key, {})):
             del cache[self.__name__]
 
